#!/usr/bin/env python3
from quickspider.core.leafpool import LeafPool
import logging


class Controller:
    def __init__(self, header):
        self.logger = logging.getLogger("Controller")
        self.logger.setLevel(logging.DEBUG)
        self._header = header
        self._leaf_pool = LeafPool(header)
        self._all_nodes = self._scan_all_nodes()
        self._register_node()
        self._lookup_collect = []
        self._status = "ACTIVATED"

    def __repr__(self):
        return f"{self._all_nodes}"

    def _init_nodes(self):
        self.logger.debug("init nodes")
        for node in self._all_nodes:
            node.init()

    def _post_nodes(self):
        self.logger.debug("post nodes")
        for node in self._all_nodes:
            node.post()

    def _register_node(self):
        self.logger.debug("register nodes")
        for node in self._all_nodes:
            node._controller = self

    def walk(self, node):
        if not node:
            return
        self.logger.debug("scan nodes")
        for child in node._child:
            self.walk(child)

    def _scan_all_nodes(self):
        nodes = []

        def walk(node):
            if not node:
                return
            nodes.append(node)
            for child in node._child:
                walk(child)

        walk(self._header)
        return nodes

    def _clear_lookup_collect(self):
        self.logger.debug("clear lookup_collect")
        self._lookup_collect = []

    @staticmethod
    def activate_rec_without_leaf(node):
        if node._is_leaf():
            return
        if node.activate():
            for child in node._child:
                Controller.activate_rec_without_leaf(child)
        return

    def activate(self):
        self.logger.debug("activate")
        if self._lookup_collect:
            for node in self._lookup_collect:
                Controller.activate_rec_without_leaf(node)
            self._clear_lookup_collect()
            return
        Controller.activate_rec_without_leaf(self._header[0])

    def loop(self):
        self.logger.debug("loop")
        self._leaf_pool.loop()

    def start(self):
        self.logger.debug("start")
        self._init_nodes()
        try:
            while True:
                # self.walk(self._header)
                self.activate()
                self.loop()
                self.lookup()
                if self._status == "FINISHED":
                    break
        except Exception as e:
            print(e)
        finally:
            self._post_nodes()

    def lookup(self):
        self.logger.debug("lookup")
        self._lookup_collect = []
        for leaf in self._leaf_pool:
            self._lookup(leaf)

    def _lookup(self, node):
        if not node:
            return
        self.logger.debug(f"[LOOKUP] -> {node._name}")
        if node._is_header():
            self._status = "FINISHED"
            return
        if node in self._lookup_collect:
            return
        if node._is_leaf():
            self._lookup(node._parent)
            return
        _not_empty = node._not_empty()
        if _not_empty:
            self._lookup_collect.append(node)
        else:
            self._lookup(node._parent)
            return



#!/usr/bin/env python3
class LeafPool:
    def __init__(self, header):
        self._pool = []
        self._scan_leaves(header)

    def __getitem__(self, key):
        return self._pool[key]

    def _add_leaf(self, node):
        self._pool.append(node)

    def _scan_leaves(self, node):
        if not node:
            return
        if node._is_leaf():
            self._add_leaf(node)
            return
        for child in node._child:
            self._scan_leaves(child)

    def loop(self):
        num = len(self._pool)
        pool = self._pool
        index = 0
        while True:
            one_leaf = pool[index]
            success = one_leaf.activate()
            if not success and index == (num-1):
                break
            index += 1
            if index == num:
                index = 0
        return



#!/usr/bin/env python3
from collections.abc import Iterable
from quickspider.core.utils import MyIterator, ProcessExceptionEmpty
from quickspider.core.error import InvalidInputError
from quickspider.core.controller import Controller
import logging


class BaseNode:
    def __init__(self, _name):
        self._name = _name
        self._input = None
        self._middle = None
        self._output = None
        self._child = []
        self._parent = None
        self._controller = None
        self._status = "EMPTY"
        self.logger = logging.getLogger(self._name)
        self.logger.setLevel(logging.DEBUG)

    def init(self):
        self.logger.debug("init")
        pass

    def post(self):
        self.logger.debug("post")
        pass

    def add_child(self, node):
        if isinstance(node, Iterable):
            for n in node:
                self._add_child(n)
            return n
        else:
            self._add_child(node)
            return node

    def __repr__(self):
        return f"""
        name: {self._name}
        status: {self._status}
        _input: {self._input}
        middle: {self._middle}
        output: {self._output}
"""

    def __getitem__(self, key):
        return self._child[key]

    def _set_input(self, _input):
        self.logger.debug("set input to {_input}")
        self._input = _input

    def _add_child(self, node):
        self._child.append(node)
        node._parent = self

    def _clear_input(self):
        self.logger.debug("clear input")
        self._input = None

    def _clear_output(self):
        self.logger.debug("clear output")
        self._output = None

    def _clear_middle(self):
        self.logger.debug("clear middle")
        self._middle = None

    def _is_leaf(self):
        return not self._child

    def _is_header(self):
        return not self._parent

    def _not_empty(self):
        _not_empty = bool(self._middle)
        if _not_empty:
            self._status = "ACTIVATED"
        else:
            self._status = "EMPTY"
        return _not_empty

    def _empty(self):
        return not self._not_empty()

    # HACK should be implement by user
    def process_input(self, _input):
        self.logger.debug("[process | input -> middle]")
        return _input

    def _input_valid(self):
        """默认的检验输入合法的方法"""
        if self._input is not None:
            return True
        else:
            return False

    def _slot_input_to_middle(self):
        if not self._input_valid():
            raise InvalidInputError
        _input = self._input
        # 具体实现委托给 process_input
        if isinstance(_input, ProcessExceptionEmpty):
            _middle = _input
        else:
            try:
                _middle = self.process_input(_input)
            except Exception as e:
                # print(e)
                _middle = ProcessExceptionEmpty()
        self._middle = _middle
        self._clear_input()
        self._not_empty()

    def _slot_fill_middle(self):
        if self._status == "EMPTY":
            self.logger.debug("activate: [EMPTY | input -> middle]")
            try:
                self._slot_input_to_middle()
            except Exception as e:
                self.logger.debug(f"activate: [EMPTY {self._name} | input -> middle] + failed with {e}")
                raise e

    def _slot_middle_to_output(self):
        self.logger.debug("activate: [middle -> output]")
        one_piece = self._middle
        self._output = one_piece
        self._clear_middle()
        self._not_empty()

    def _slot_output_to_next(self):
        if self._is_leaf():
            self.logger.debug("activate: [LEAF | output -> child]")
            _output = self._slot_output_to_child_leaf()
        else:
            self.logger.debug("activate: [output -> child]")
            if isinstance(self._output, ProcessExceptionEmpty):
                return
            _output = self._slot_output_to_child()
        return _output

    def _slot_output_to_child(self):
        _output = self._output
        for child in self._child:
            child._set_input(_output)
        self._clear_output()
        self._not_empty()
        return _output

    def _slot_output_to_child_leaf(self):
        _output = self._output
        print(_output)
        print("")
        self._clear_output()
        return _output

    # HACK CORE
    def activate(self):
        self.logger.debug("activate: start")
        # 如果需要编写抽象类，如IteratorNode，则必须实现_empty()
        try:
            self._slot_fill_middle()
            self._slot_middle_to_output()
            _output = self._slot_output_to_next()
        except Exception as e:
            # print(e)
            return False
        return True, _output


class IterableNode(BaseNode):
    def __init__(self, _name):
        super().__init__(_name)
        self._middle = MyIterator([])

    def _not_empty(self):
        _not_empty = self._middle.not_empty()
        if _not_empty:
            self._status = "ACTIVATED"
        else:
            self._status = "EMPTY"
        return _not_empty

    def _slot_input_to_middle(self):
        assert self._input is not None
        _input = self._input
        _middle = self.process_input(_input)
        self._middle = MyIterator(_middle)
        self._clear_input()
        self._not_empty()

    def _slot_middle_to_output(self):
        one_piece = self._fetch_one_piece()
        self._output = one_piece
        self._not_empty()

    def _fetch_one_piece(self):
        if not self._not_empty():
            return None
        return self._middle.fetch_one_piece()

    # HACK test
    def process_input(self, _input):
        return (_input+f"_{i}_by_[{self._name}]" for i in range(5))


if __name__ == "__main__":
    header = BaseNode("header")
    node_1 = BaseNode("node_1")
    node_2 = BaseNode("node_2")
    node_3 = BaseNode("node_3")
    node_4 = BaseNode("node_4")
    header.add_child(node_1).add_child(node_2).add_child(node_3).add_child(node_4)
    node_1._set_input("测试")
    controller = Controller(header)



#!/usr/bin/env python3
import inspect
import quickspider.core.Node
import quickspider.core.nodes
from quickspider.core.Node import BaseNode
from collections import deque

# scan the node class in nodes
node_class = {"basenode": quickspider.core.Node.BaseNode}
for name, obj in inspect.getmembers(quickspider.core.nodes):
    if inspect.isclass(obj) and issubclass(obj, quickspider.core.Node.BaseNode):
        node_class[name.lower()] = obj

class Builder:
    node_class = node_class
    def __init__(self):
        self._nodes = deque()

    def _add_node(self, _node):
        self._nodes.append(_node)

    def node_build(self, _name, _config):
        node_class_name = _config.pop("type").lower()
        node_class = Builder.get_class_from_name(node_class_name)
        config = {f"_{k}": v for k, v in _config.items()}
        name = _name.split(".")[-1]
        node = node_class(_name=name, **config)
        self._add_node(node)
        return node

    def link(self):
        header = BaseNode("header")
        node = header
        while self._nodes:
            next_node = self._nodes.popleft()
            node = node.add_child(next_node)
        return header

    @staticmethod
    def get_class_from_name(node_class_name):
        return Builder.node_class.get(node_class_name, BaseNode)


if __name__ == "__main__":
    b = Builder()



from quickspider.core.Node import BaseNode, IterableNode
from scrapy import Selector
from requests import Session
from termcolor import colored
import openpyxl
import csv
import json
import os
s = Session()
s.headers = {"User-Agent": "Google-Bot"}


#-------------> 网络节点 <--------------
class SessionNode(BaseNode):
    def __init__(self, _name, _session=s):
        super().__init__(_name)
        self._session = _session


class GetNode(SessionNode):
    def process_input(self, _input_url):
        try:
            resp = self._session.get(_input_url)
        except Exception as e:
            print(colored(str(e), "red"))
            resp = None
        return resp


class PostNode(SessionNode):
    def _set_data(self, _data: dict=None):
        if not _data:
            _data = {}
        self._data = _data

    def process_input(self, _input_url):
        try:
            resp = self._session.post(_input_url, json=self._data)
        except Exception as e:
            print(colored(str(e), "red"))
            resp = None
        return resp


#-------------> 解析节点 <--------------
class ParserNode(BaseNode):
    def __init__(self, _name, _mode):
        super().__init__(_name)
        self._set_mode(_mode)

    def _set_mode(self, _mode="dom"):
        self._mode = _mode

    def process_input(self, _response):
        if self._mode == "dom":
            return Selector(text=_response.text)
        if self._mode == "json":
            return _response.json()


class ParserDomNode(BaseNode):
    def __init__(self, _name, _mode="css", _parser=""):
        super().__init__(_name)
        self._set_mode(_mode)
        self._set_parser(_parser)

    def _set_mode(self, _mode="css"):
        self._mode = _mode

    def _set_parser(self, _parser):
        self._parser = _parser

    def process_input(self, _response):
        selector = _response
        if self._mode == "css":
            result = selector.css(self._parser)
        elif self._mode == "xpath":
            result = selector.xpath(self._parser)
        return result


class ParserJsonNode(BaseNode):
    pass


#-------------> 组合节点 <--------------
class ExtractNode(IterableNode):
    def process_input(self, _input_parser):
        result = _input_parser.getall()
        return (i for i in result)


class ConcatNode(BaseNode):
    def process_input(self, _input_parser):
        result = "".join(_input_parser.getall())
        return result


#-------------> 输出测试节点 <--------------
class PrintNode(BaseNode):
    def process_input(self, _input):
        print(colored(_input, "green"))
        return _input


class PageNode(IterableNode):
    def __init__(self, _name, _input, _start, _stop, _step=1):
        super().__init__(_name)
        self._set_input(_input)
        self._info = (_start, _stop, _step)

    def process_input(self, _input):
        _start, _stop, _step = self._info
        return (_input.format(page=p) for p in range(_start, _stop, _step))


#-------------> 读写节点 <--------------
FileReaderNode = IterableNode

class CsvReaderNode(FileReaderNode):
    def __init__(self, _name, _file, _column):
        super().__init__(_name)
        self._set_input(_file)
        self._column = _column

    def process_input(self, _input):
        if not os.path.exists(_input):
            raise Exception("Not Exist")
        with open(_input, mode="rt") as csvfile:
            csvdictreader = csv.DictReader(csvfile)
            for row in csvdictreader:
                yield row[self._column]
        

class ExcelReaderNode(FileReaderNode):
    def __init__(self, _name, _file, _column):
        super().__init__(_name)
        self._set_input(_file)
        self._column = _column

    def process_input(self, _input):
        wb = openpyxl.load_workbook(_input)
        # select active sheet by default
        sheet = wb.active
        for col in range(1, sheet.max_column + 1):
            if sheet.cell(1, col).value == self._column:
                col_index = col
                break
        for row in range(2, sheet.max_rol + 1):
            cell = sheet.cell(row=row, column=col_indx)
            yield cell.value


class LineReaderNode(FileReaderNode):
    def __init__(self, _name, _file):
        super().__init__(_name)
        self._set_input(_file)

    def process_input(self, _input):
        with open(_input, "rt", encoding="utf8") as f:
            for line in f:
                yield line


class JsonWriterNode(BaseNode):
    def __init__(self, _name, _file):
        super().__init__(_name)
        self._file = _file
        self._container = []

    def init(self):
        if os.path.exists(self._file):
            os.remove(self._file)

    def process_input(self, _input):
        self._container.append(_input)
        return _input

    def post(self):
        with open(self._file, "wt", encoding="utf8") as f:
            json.dump(self._container, f, ensure_ascii=False, indent=4)

class CsvWriterNode(BaseNode):
    def __init__(self, _name, _file):
        super().__init__(_name)
        self._file = _file

    def init(self):
        if os.path.exists(self._file):
            os.remove(self._file)
        self._file_handel = open(self._file, "wt", encoding="utf8")
        self._csv_writer = csv.writer(self._file_handel)
        self._csv_writer.writerow(["result"])

    def process_input(self, _input):
        self._csv_writer.writerow([_input])
        return _input

    def post(self):
        self._file_handel.close()

# add your own node 




#!/usr/bin/env python3
from itertools import chain
from collections.abc import Iterator


class MyIterator:
    def __init__(self, iterator):
        if not isinstance(iterator, Iterator):
            iterator = iter(iterator)
        self._container = iterator

    def fetch_one_piece(self):
        try:
            one_piece = next(self._container)
        except StopIteration:
            one_piece = None
        return one_piece

    def not_empty(self):
        one_piece = self.fetch_one_piece()
        if one_piece:
            self._container = chain([one_piece], self._container)
            return True
        return False


class ProcessExceptionEmpty:
    pass



#!/usr/bin/env python3

from distutils.core import setup

with open("./README.md", "rt") as f:
    long_description = f.read()

setup(
        name='quickspider',
        version='0.1.2.3',
        description='help you build a small spider quickly.',
        long_description="see github.",
        author='Ivory',
        author_email='1070642565@qq.com',
        url='https://github.com/tiwe0/quickspider.git',
        packages=['quickspider', 'quickspider.core', 'quickspider.template',
            'quickspider.test'],
        package_dir={'quickspider': 'quickspider'},
        package_data={'quickspider': ['template/*.toml']},
        scripts=['quickspider/bin/quickspidercommand']
        )



import unittest
import requests
from quickspider.core.nodes import (GetNode, PostNode, ParserNode,
        ParserDomNode, ExtractNode, ConcatNode, PrintNode, PageNode, CsvReaderNode,
        ExcelReaderNode, LineReaderNode, JsonWriterNode, CsvWriterNode)
from scrapy import Selector


class TestNodes(unittest.TestCase):

    def test_GetNode(self):
        getnode = GetNode("test")
        getnode._set_input("https://httpbin.org/get")
        status, resp = getnode.activate()
        self.assertTrue(status)
        self.assertEqual(resp.status_code, 200)

    def test_PostNode(self):
        postnode = PostNode("test")
        postnode._set_data({"test": "test_node"})
        status, resp = postnode.activate()
        self.assertTrue(status)
        self.assertEqual(resp.status_code, 200)

    def test_ParserNodeDom(self):
        resp = requests.get("https://baidu.com")
        parsernode = ParserNode("test", "dom")
        parsernode._set_input(resp)
        status, selector =  parsernode.activate()
        self.assertTrue(status)
        self.assertIsInstance(selector, Selector)

    def test_ParserNodeJson(self):
        data = {"test": "test_json"}
        resp = requests.post("https://httpbin.org/post", json=data)
        parsernode = ParserNode("test", "json")
        parsernode._set_input(resp)
        status, jsondata =  parsernode.activate()
        self.assertTrue(status)
        self.assertDictEqual(jsondata["data"], data)

    def get_selector(self):
        test_dom_text = """
        <body>
        <h1>Method Not Allowed</h1>
        <p>The method is not allowed for the requested URL.</p>
        </body>
        """
        test_dom = Selector(text=test_dom_text)
        parserdomnode = ParserDomNode("test", "css", "p::text")
        parserdomnode._set_input(test_dom)
        status, selector = parserdomnode.activate()
        return status, selector

    def test_ParserDomNode(self):
        status, selector = self.get_selector()
        self.assertTrue(status)
        self.assertIsInstance(selector, Selector)
        return selector

    def test_ExtractNode(self):
        _, selector = self.get_selector()
        extractnode = ExtractNode("test")
        extractnode._set_input(selector)
        status, result = extractnode.activate()
        self.assertTrue(status)
        self.assertEqual(result, ["The method is not allowed for the requested URL."])

    def test_ConcatNode(self):
        _, selector = self.get_selector()
        concatnode = ConcatNode("test")
        concatnode._set_input(selector)
        status, result = concatnode.activate()
        self.assertTrue(status)
        self.assertEqual(result, "The method is not allowed for the requested URL.")

    def test_PrintNode(self):
        printnode = PrintNode("test")
        printnode._set_input("test passed")
        status, result = printnode.activate()
        self.assertTrue(status)
        print(result)

    def test_PageNode(self):
        pagenode = PageNode("test", "test/{page}", 1, 4)
        page = 1
        status, output = pagenode.activate()
        while status:
            status, output = pagenode.activate()
            if not status:
                break
            self.assertEqual(output, f"test/{page}")
            page += 1


